import os
from uuid import uuid4
import json
import traceback
import tempfile
import logging
log = logging.getLogger()

from arcgis.widgets._mapview._raster.engine import RasterData, MatplotlibRasterEngine

class ArcPyRasterEngine(MatplotlibRasterEngine):
    def try_get_raster_data(self, image, bands) -> RasterData:
        try:
            return self._try_get_raster_data(image, bands)
        except Exception as e:
            return super().try_get_raster_data(image, bands)

    def _try_get_raster_data(self, image, bands) -> RasterData:
        from arcpy import Raster, RasterToNumPyArray 
        raster = Raster(image)
        img_data = RasterToNumPyArray(raster)
        if len(img_data.shape) == 3:
            img_data = img_data.transpose((1,2,0)).squeeze()
            if bands:
                img_data = img_data[:,:, bands].squeeze()
        extent = self._get_extent(raster)
        return RasterData(img_data = img_data,
                          extent = extent,
                          id_ = str(hash(raster.extent.JSON + raster.name)))

    def _get_extent(self, raster):
        import arcpy
        from arcpy import Raster, RasterToNumPyArray, SpatialReference
        try:
            transformations = arcpy.ListTransformations(
                raster.spatialReference, arcpy.SpatialReference(4326))
            if len(transformations) > 0:
                transformation = transformations[0]
            else:
                transformation = None
            extent = json.loads(raster.extent.projectAs(SpatialReference(4326),
                                                        transformation).JSON)
            if extent["spatialReference"]["wkid"] is not None:
                return extent
            else:
                raise Exception(f"{extent} contains invalid spatial reference")
        except Exception as e:
            log.debug("Handled exception on inferring extent w/ arcpy engine")
            log.debug(e)
            return None



import json
import ujson as _ujson
from collections import OrderedDict
from collections.abc import MutableMapping, Mapping

###########################################################################
class InsensitiveDict(MutableMapping):
    """
    A case-insensitive ``dict`` like object used to update and alter JSON

    A varients of a case-less dictionary that allows for dot and bracket notation.
    """
    #----------------------------------------------------------------------
    def __init__(self, data=None):
        self._store = OrderedDict() # 
        if data is None:
            data = {}
            
        self.update(data)     
        self._to_isd(self)
    #----------------------------------------------------------------------
    def __repr__(self):
        return str(dict(self.items()))
    #----------------------------------------------------------------------
    def __str__(self):
        return str(dict(self.items()))    
    #----------------------------------------------------------------------
    def __setitem__(self, key, value):
        if str(key).lower() in {'_store'}:
            super(InsensitiveDict, self).__setattr__(key,value)
        else:
            if isinstance(value, dict):
                self._store[str(key).lower()] = (key, InsensitiveDict(value))
            else:
                self._store[str(key).lower()] = (key, value)        
    #----------------------------------------------------------------------
    def __getitem__(self, key):
        return self._store[str(key).lower()][1]
    #----------------------------------------------------------------------
    def __delitem__(self, key):
        del self._store[str(key).lower()]
    #----------------------------------------------------------------------
    def __getattr__(self, key):
        if str(key).lower() in {"_store"}:
            return self._store
        else:
            return self._store[str(key).lower()][1]
    #----------------------------------------------------------------------
    def __setattr__(self, key, value):
        if str(key).lower() in {'_store'}:
            super(InsensitiveDict, self).__setattr__(key,value)
        else:
            if isinstance(value, dict):
                self._store[str(key).lower()] = (key, InsensitiveDict(value))
            else:
                self._store[str(key).lower()] = (key, value)
    #----------------------------------------------------------------------
    def __dir__(self):
        return self.keys()    
    #----------------------------------------------------------------------
    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in self._store.values())
    #----------------------------------------------------------------------
    def __len__(self):
        return len(self._store)
    #----------------------------------------------------------------------
    def _lower_items(self):
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lowerkey, keyval[1])
            for (lowerkey, keyval)
            in self._store.items()
        )
    #----------------------------------------------------------------------
    def __setstate__(self, d):
        """unpickle support """
        self.__dict__.update(InsensitiveDict(d).__dict__)
        self = InsensitiveDict(d)
    #----------------------------------------------------------------------
    def __getstate__(self):
        """ pickle support """
        return _ujson.loads(self.json)    
    #----------------------------------------------------------------------
    @classmethod
    def from_dict(cls, o):
        """Converts dict to a InsensitiveDict"""
        return cls(o)
    #----------------------------------------------------------------------
    @classmethod
    def from_json(cls, o):
        """Converts JSON string to a InsensitiveDict"""
        if isinstance(o, str):
            o = _ujson.loads(o)
            return InsensitiveDict(o)
        return InsensitiveDict.from_dict(o)    
    #----------------------------------------------------------------------
    def __eq__(self, other):
        if isinstance(other, Mapping):
            other = InsensitiveDict(other)
        else:
            return NotImplemented
        return dict(self._lower_items()) == dict(other._lower_items())
    #----------------------------------------------------------------------
    def copy(self):
        return InsensitiveDict(self._store.values())
    #---------------------------------------------------------------------
    def _to_isd(self, data):
        """converts a dictionary from InsensitiveDict to a dictionary"""
        for k,v in data.items():
            if isinstance(v, (dict, InsensitiveDict)):
                data[k] = self._to_isd(v)
            elif isinstance(v, (list, tuple)):
                l = []
                for i in v:
                    if isinstance(i, dict):
                        l.append(InsensitiveDict(i))
                    else:
                        l.append(i)
                if isinstance(v, tuple):
                    l = tuple(l)
                data[k] = l
        return data
    #---------------------------------------------------------------------
    def _json(self):
        """converts an InsensitiveDict to a dictionary"""
        d = {}
        for k, v in self.items():
            if isinstance(v, InsensitiveDict):
                d[k] = v._json()
            elif type(v) in (list, tuple):
                l = []
                for i in v:
                    if isinstance(i, InsensitiveDict):
                        l.append(i._json())
                    else:
                        l.append(i)
                if type(v) is tuple:
                    v = tuple(l)
                else:
                    v = l
                d[k] = v
            else:
                d[k] = v # not list, tuple, or InsensitiveDict
        return d
    #----------------------------------------------------------------------
    @property
    def json(self):
        """returns the value as JSON String"""
        o = self._json()#dict(self.copy())
        return _ujson.dumps(dict(o))




import os
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap

########################################################################
class SecurityManager(object):
    """
    This resource is a container for all resources and operations
    pertaining to security in your ArcGIS Mission Server site.
    """
    _url = None
    _gis = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<SecurityManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def _modify_server_role(self, role):
        """
        Allows for the modification of the server role from federated to standalone.
        
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        role                   Required String. The value that determines if the server is federated
                               or standalone.  This allowed values are: "FEDERATED_SERVER" or 
                               "STANDALONE_SERVER".
        ==================     ====================================================================
        
        :returns: Dict
        
        """
        url = self._url + "/config/changeServerRole"
        function = None  
        if function is None:
            function = ""
        params = {'f' : 'json',
                  'serverRole' : role,
                  'serverFunction' : function}
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    @property
    def configuration(self):
        """
        This resource returns the currently active security configuration
        of your ArcGIS Mission Server site. A security configuration
        involves the following pieces of information, and can be modified
        using the setter.

        **Configuration Properties**

        ===========================     ====================================================================
        **Key**                         **Description**
        ---------------------------     --------------------------------------------------------------------
        allowedAdminAccessIPs           (Allowed Administration Access IPs). A comma separated list of client machine IP addresses that are allowed access to ArcGIS Server. This can be used as an additional security measure to prevent unauthorized access to the site.
        ---------------------------     --------------------------------------------------------------------
        allowDirectAccess               (Allow direct administrator access). A boolean that indicates if a user with administrator privileges can access the server through port 6080. If true, all users with administrative access are allowed to access the Administrator Directory and ArcGIS Server Manager through port 6080. If false, users in the identity store are blocked from accessing the server through port 6080. This forces users to access the site through ArcGIS Web Adaptor. The default value is true.
    
                                        Before disabling administrative access on port 6080, ArcGIS Server must be configured to use web tier authentication (WEB_ADAPTOR) and at least one user in the identity store must have administrator privileges to the site. The primary site administrator account will still be able to administer the site through port 6080.

                                        To fully disable access on port 6080, you can optionally disabled the primary site administrator account. If ArcGIS Server Manager becomes unavailable or the web server is unable to authenticate users that have administrator privileges, you will be unable to administer your site. To recover from this site, re-enable the primary site administrator account and connect to the site through port 6080 with this account.

        ---------------------------     --------------------------------------------------------------------
        authenticationMode              (Authentication Mode). Specifies the authentication mode used by ArcGIS Server. When ArcGIS Server is federated with Portal for ArcGIS, this property can be included and set to ARCGIS_PORTAL_TOKEN. The default value is ARCGIS_TOKEN.

                                        Values: ARCGIS_TOKEN | ARCGIS_PORTAL_TOKEN | WEB_ADAPTOR_AUTHENTICATION
        ---------------------------     --------------------------------------------------------------------
        authenticationTier              (Authentication Tier). The tier at which requests to access GIS services will be authenticated. It is recommended that you do not modify these values using the Administrator Directory. Instead, use ArcGIS Server Manager to configure web tier authentication or use the Portal for ArcGIS website to federate ArcGIS Server with your portal.

                                        Values: WEB_ADAPTOR | GIS_SERVER | ARCGIS_PORTAL
        ---------------------------     --------------------------------------------------------------------
        HSTSEnabled                     (HSTS Enabled). A boolean that indicates if HTTP Strict Transport Security (HSTS) is being used by the site. See Enforce strict HTTPS communication for more information. In order for this property to be enabled, the Protocol property must already be set to use HTTPS only.
        ---------------------------     --------------------------------------------------------------------
        httpEnabled                     (HTTP Enabled). A boolean that indicates if the site is accessible over HTTP.
        ---------------------------     --------------------------------------------------------------------
        Protocol                        (Protocol). Specifies the HTTP protocol to be used for communication to and from the ArcGIS Server site. If set to HTTP, all communication to and from the site will be over HTTP, with HTTPS communication being unavailable. If HTTP_AND_HTTPS is set, users and client may use either HTTP or HTTPS to connect to the site. If HTTPS, all communication to and from the site will be over HTTPS. Any call made using HTTP will be redirected to use HTTPS instead.
                                        When you initially create your ArcGIS Server site, all communication in the site is sent over HTTP, which is not secure. This means that your credentials sent over an internal network or the Internet are not encrypted and can be intercepted. To prevent the interception of any communication, it's recommended that you configure ArcGIS Server and ArcGIS Server Manager (if installed) to enforce Secure Sockets Layer (SSL). When you initially create your site, you'll see a warning-level message in the logs recommending that you update the communication protocol of your site to use SSL.

                                        Values: HTTP | HTTP_AND_HTTPS | HTTPS
        ---------------------------     --------------------------------------------------------------------
        roleStoreConfig                 (Role Store). Connection information about the currently active role store.
                                    
        ---------------------------     --------------------------------------------------------------------
        securityEnabled                 (Security Enabled). A boolean that indicates if security is enabled for any GIS service. The default value is true.
        ---------------------------     --------------------------------------------------------------------
        sslEnabled                      (SSL Enabled). A boolean that indicates if the site is accessible over HTTPS (SSL). The default value is false.
        ---------------------------     --------------------------------------------------------------------
        userStoreConfig                 (User Store). Connection information about the currently active user store.                                        
        ---------------------------     --------------------------------------------------------------------
        virtualDirsSecurityEnabled      (Virtual Directories Security Enabled). A boolean that indicates if the server's virtual directories are secured and require authentication. If true, accessing the content in the arcgisoutput, arcgisjobs, and arcgisinput directories over HTTP will require user authentication. This will negatively impact performance. The default value is false.
        ---------------------------     --------------------------------------------------------------------
        portalProperties                (Portal Properties). Specified when federating ArcGIS Server with Portal for ArcGIS. See the Portal properties for more information.
        ===========================     ====================================================================



        **Portal Properties**

        ===========================     ====================================================================
        **Portal Keys**                 **Description**
        ---------------------------     --------------------------------------------------------------------
        portalMode                      Must be set as ARCGIS_PORTAL_FEDERATION.
        ---------------------------     --------------------------------------------------------------------
        portalSecretKey                 The key obtained after federating ArcGIS Server with Portal for ArcGIS.
        ---------------------------     --------------------------------------------------------------------
        portalURL                       The URL of Portal for ArcGIS.
        ---------------------------     --------------------------------------------------------------------
        referer                         The referer specified when generating the token.
        ---------------------------     --------------------------------------------------------------------
        serverId                        The ID of the server federated with the portal.
        ---------------------------     --------------------------------------------------------------------
        serverURL                       External URL of the server federated with the portal in the following format:
        ---------------------------     --------------------------------------------------------------------
        token                           A token obtained from Portal for ArcGIS for use by ArcGIS Server for initial validation.
        ===========================     ====================================================================
        


        

        """
        url = self._url + "/config"
        params = {'f' : 'json'}
        return self._con.get(url, params)
    #----------------------------------------------------------------------
    @configuration.setter
    def configuration(self, settings):
        """
        This resource returns the currently active security configuration
        of your ArcGIS Mission Server site. A security configuration
        involves the following pieces of information, and can be modified
        using the setter.

        **Configuration Properties**

        ===========================     ====================================================================
        **Key**                         **Description**
        ---------------------------     --------------------------------------------------------------------
        allowedAdminAccessIPs           (Allowed Administration Access IPs). A comma separated list of client machine IP addresses that are allowed access to ArcGIS Server. This can be used as an additional security measure to prevent unauthorized access to the site.
        ---------------------------     --------------------------------------------------------------------
        allowDirectAccess               (Allow direct administrator access). A boolean that indicates if a user with administrator privileges can access the server through port 6080. If true, all users with administrative access are allowed to access the Administrator Directory and ArcGIS Server Manager through port 6080. If false, users in the identity store are blocked from accessing the server through port 6080. This forces users to access the site through ArcGIS Web Adaptor. The default value is true.
    
                                        Before disabling administrative access on port 6080, ArcGIS Server must be configured to use web tier authentication (WEB_ADAPTOR) and at least one user in the identity store must have administrator privileges to the site. The primary site administrator account will still be able to administer the site through port 6080.

                                        To fully disable access on port 6080, you can optionally disabled the primary site administrator account. If ArcGIS Server Manager becomes unavailable or the web server is unable to authenticate users that have administrator privileges, you will be unable to administer your site. To recover from this site, re-enable the primary site administrator account and connect to the site through port 6080 with this account.

        ---------------------------     --------------------------------------------------------------------
        authenticationMode              (Authentication Mode). Specifies the authentication mode used by ArcGIS Server. When ArcGIS Server is federated with Portal for ArcGIS, this property can be included and set to ARCGIS_PORTAL_TOKEN. The default value is ARCGIS_TOKEN.

                                        Values: ARCGIS_TOKEN | ARCGIS_PORTAL_TOKEN | WEB_ADAPTOR_AUTHENTICATION
        ---------------------------     --------------------------------------------------------------------
        authenticationTier              (Authentication Tier). The tier at which requests to access GIS services will be authenticated. It is recommended that you do not modify these values using the Administrator Directory. Instead, use ArcGIS Server Manager to configure web tier authentication or use the Portal for ArcGIS website to federate ArcGIS Server with your portal.

                                        Values: WEB_ADAPTOR | GIS_SERVER | ARCGIS_PORTAL
        ---------------------------     --------------------------------------------------------------------
        HSTSEnabled                     (HSTS Enabled). A boolean that indicates if HTTP Strict Transport Security (HSTS) is being used by the site. See Enforce strict HTTPS communication for more information. In order for this property to be enabled, the Protocol property must already be set to use HTTPS only.
        ---------------------------     --------------------------------------------------------------------
        httpEnabled                     (HTTP Enabled). A boolean that indicates if the site is accessible over HTTP.
        ---------------------------     --------------------------------------------------------------------
        Protocol                        (Protocol). Specifies the HTTP protocol to be used for communication to and from the ArcGIS Server site. If set to HTTP, all communication to and from the site will be over HTTP, with HTTPS communication being unavailable. If HTTP_AND_HTTPS is set, users and client may use either HTTP or HTTPS to connect to the site. If HTTPS, all communication to and from the site will be over HTTPS. Any call made using HTTP will be redirected to use HTTPS instead.
                                        When you initially create your ArcGIS Server site, all communication in the site is sent over HTTP, which is not secure. This means that your credentials sent over an internal network or the Internet are not encrypted and can be intercepted. To prevent the interception of any communication, it's recommended that you configure ArcGIS Server and ArcGIS Server Manager (if installed) to enforce Secure Sockets Layer (SSL). When you initially create your site, you'll see a warning-level message in the logs recommending that you update the communication protocol of your site to use SSL.

                                        Values: HTTP | HTTP_AND_HTTPS | HTTPS
        ---------------------------     --------------------------------------------------------------------
        roleStoreConfig                 (Role Store). Connection information about the currently active role store.
                                    
        ---------------------------     --------------------------------------------------------------------
        securityEnabled                 (Security Enabled). A boolean that indicates if security is enabled for any GIS service. The default value is true.
        ---------------------------     --------------------------------------------------------------------
        sslEnabled                      (SSL Enabled). A boolean that indicates if the site is accessible over HTTPS (SSL). The default value is false.
        ---------------------------     --------------------------------------------------------------------
        userStoreConfig                 (User Store). Connection information about the currently active user store.                                        
        ---------------------------     --------------------------------------------------------------------
        virtualDirsSecurityEnabled      (Virtual Directories Security Enabled). A boolean that indicates if the server's virtual directories are secured and require authentication. If true, accessing the content in the arcgisoutput, arcgisjobs, and arcgisinput directories over HTTP will require user authentication. This will negatively impact performance. The default value is false.
        ---------------------------     --------------------------------------------------------------------
        portalProperties                (Portal Properties). Specified when federating ArcGIS Server with Portal for ArcGIS. See the Portal properties for more information.
        ===========================     ====================================================================



        **Portal Properties**

        ===========================     ====================================================================
        **Portal Keys**                 **Description**
        ---------------------------     --------------------------------------------------------------------
        portalMode                      Must be set as ARCGIS_PORTAL_FEDERATION.
        ---------------------------     --------------------------------------------------------------------
        portalSecretKey                 The key obtained after federating ArcGIS Server with Portal for ArcGIS.
        ---------------------------     --------------------------------------------------------------------
        portalURL                       The URL of Portal for ArcGIS.
        ---------------------------     --------------------------------------------------------------------
        referer                         The referer specified when generating the token.
        ---------------------------     --------------------------------------------------------------------
        serverId                        The ID of the server federated with the portal.
        ---------------------------     --------------------------------------------------------------------
        serverURL                       External URL of the server federated with the portal in the following format:
        ---------------------------     --------------------------------------------------------------------
        token                           A token obtained from Portal for ArcGIS for use by ArcGIS Server for initial validation.
        ===========================     ====================================================================
        


        """
        url = self._url + "/config/update"
        params = {'f' : 'json'}
        current = dict(self.configuration)
        for k, v in settings.items():
            if k in current:
                params[k] = settings[k]
                current.pop(k)
        params.update(current)
        return self._con.post(url, params)



import random
import tempfile
import warnings
import sys
import math
import os
from pathlib import Path

import arcgis
from arcgis.features import FeatureLayer

HAS_FASTAI = True
try:
    from fastai.tabular import TabularList
    from fastai.tabular import TabularDataBunch
    from fastai.tabular.transform import FillMissing, Categorify, Normalize
    from fastai.tabular import cont_cat_split, add_datepart
    from fastai.data_block import ItemLists, CategoryList, FloatList
    from .._utils.TSData import TimeSeriesList, To3dTensor
    from fastai.data_block import DatasetType
    import torch
except Exception as e:
    HAS_FASTAI = False

HAS_NUMPY = True
try:
    import numpy as np
except:
    HAS_NUMPY = False

HAS_SK_LEARN = True
try:
    from sklearn.pipeline import make_pipeline
    from sklearn.compose import make_column_transformer
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import Normalizer, LabelEncoder, MinMaxScaler
except:
    HAS_SK_LEARN = False


class DummyTransform(object):
    def __int__(self):
        pass

    def fit_transform(self, x):
        return x

    def inverse_trasnform(self, x):
        return x


class TabularDataObject(object):
    _categorical_variables = []
    _continuous_variables = []
    dependent_variables = []

    @classmethod
    def prepare_data_for_layer_learner(
        cls,
        input_features,
        dependent_variable,
        feature_variables=None,
        raster_variables=None,
        date_field=None,
        distance_feature_layers=None,
        procs=None,
        val_split_pct=0.1,
        seed=42,
        batch_size=64,
        index_field=None,
        column_transforms_mapping=None
    ):

        if not HAS_FASTAI:
            return

        feature_variables = feature_variables if feature_variables else []
        raster_variables = raster_variables if raster_variables else []

        tabular_data = cls()
        tabular_data._dataframe, tabular_data._field_mapping = TabularDataObject._prepare_dataframe_from_features(
            input_features,
            dependent_variable,
            feature_variables,
            raster_variables,
            date_field,
            distance_feature_layers,
            index_field
        )

        if input_features is None:
            tabular_data._is_raster_only = True
        else:
            tabular_data._is_raster_only = False

        tabular_data._dataframe = tabular_data._dataframe.reindex(sorted(tabular_data._dataframe.columns), axis=1)

        tabular_data._categorical_variables = tabular_data._field_mapping['categorical_variables']
        tabular_data._continuous_variables = tabular_data._field_mapping['continuous_variables']
        tabular_data._dependent_variable = tabular_data._field_mapping['dependent_variable']
        tabular_data._index_data = tabular_data._field_mapping['index_data']
        tabular_data._index_field = index_field

        tabular_data._procs = procs
        tabular_data._column_transforms_mapping = column_transforms_mapping
        tabular_data._val_split_pct = val_split_pct
        tabular_data._bs = batch_size
        tabular_data._seed = seed

        validation_indexes = []
        if tabular_data._dependent_variable:
            random.seed(seed)
            validation_indexes = random.sample(range(len(tabular_data._dataframe)), round(val_split_pct * len(tabular_data._dataframe)))
            tabular_data._validation_indexes = validation_indexes

        tabular_data._training_indexes = list(set([i for i in range(len(tabular_data._dataframe))]) - set(validation_indexes))
        if not tabular_data._dependent_variable:
            tabular_data._validation_indexes = list(set([i for i in range(len(tabular_data._dataframe))]))

        tabular_data._is_empty = False

        if tabular_data._dependent_variable:
            tabular_data._is_classification = tabular_data._is_classification()
        else:
            tabular_data._is_classification = True

        tabular_data.path = Path(os.getcwd())
        return tabular_data

    @staticmethod
    def _min_of(values_list):
        return_values = []
        for values in values_list:
            if len(values) == 0:
                return_values.append(0)
                continue

            if len(values) == 1:
                return_values.append(values[0])
                continue

            min_value = values[0]
            for value in values:
                if value < min_value:
                    min_value = value
            return_values.append(min_value)

        return return_values

    @staticmethod
    def _max_of(values_list):
        return_values = []
        for values in values_list:
            if len(values) == 0:
                return_values.append(0)
                continue

            return_values.append(max(values))

        return return_values

    @staticmethod
    def _mean_of(values_list):
        return_values = []
        for values in values_list:
            if len(values) == 0:
                return_values.append(0)
                continue

            return_values.append(sum(values) / len(values))

        return return_values

    @staticmethod
    def _majority_of(values_list):
        return_values = []
        for values in values_list:
            if len(values) == 0:
                return_values.append(0)
                continue

            return_values.append(max(values, key=values.count))

        return return_values

    @staticmethod
    def _minority_of(values_list):
        return_values = []
        for values in values_list:
            if len(values) == 0:
                return_values.append(0)
                continue

            return_values.append(min(values, key=values.count))

        return return_values

    @staticmethod
    def _sum_of(values_list):
        return_values = []
        for values in values_list:
            if len(values) == 0:
                return_values.append(0)
                continue

            return_values.append(sum(values))

        return return_values

    @staticmethod
    def _std_dev_of(values_list):
        import statistics
        return_values = []

        for values in values_list:
            if len(values) == 0:
                return_values.append(0)
                continue

            return_values.append(statistics.stdev(values))

        return return_values

    @staticmethod
    def _variety(values_list):
        return_values = []
        for values in values_list:
            return_values.append(len(list(set(values))))

        return return_values

    @staticmethod
    def _get_calc(raster_type, calc_type):
        calc_type = calc_type.lower()

        cont_mapping = {
            'min': TabularDataObject._min_of,
            'max': TabularDataObject._max_of,
            'mean': TabularDataObject._mean_of,
            'majority': TabularDataObject._majority_of,
            'minority': TabularDataObject._minority_of,
            'std_dev': TabularDataObject._std_dev_of,
            'sum': TabularDataObject._sum_of,
            'variety': TabularDataObject._variety
        }

        cat_mapping = {
            'majority': TabularDataObject._majority_of,
            'minority': TabularDataObject._minority_of,
            'variety': TabularDataObject._variety
        }

        if raster_type:
            return cat_mapping.get(calc_type, TabularDataObject._majority_of)
        else:
            return cont_mapping.get(calc_type, TabularDataObject._mean_of)

    def _prepare_validation_databunch(self, dataframe):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            kwargs_variables = {'num_workers': 0} if sys.platform == 'win32' else {}
            # kwargs_variables['tfm_y'] = True
            fm = FillMissing(self._categorical_variables, self._continuous_variables)
            fm.add_col = False
            fm(dataframe)
            databunch_half = TabularList.from_df(
                dataframe,
                path=tempfile.NamedTemporaryFile().name,
                cat_names=self._categorical_variables,
                cont_names=self._continuous_variables,
                procs=[Categorify, Normalize]
            ).split_by_idx([i for i in range(int(len(dataframe)/2))]).label_empty().databunch(**kwargs_variables)

            databunch_second_half = TabularList.from_df(
                dataframe,
                path=tempfile.NamedTemporaryFile().name,
                cat_names=self._categorical_variables,
                cont_names=self._continuous_variables,
                procs=[Categorify, Normalize]
            ).split_by_idx([i for i in range(int(len(dataframe)/2), len(dataframe))]).label_empty().databunch(**kwargs_variables)

        return databunch_half, databunch_second_half

    @property
    def _is_unsupervised(self):
        if not self._dependent_variable:
            return True

        return False

    def _is_classification(self):
        if self._is_empty:
            return True

        if not HAS_NUMPY:
            raise Exception("This module requires numpy.")

        df = self._dataframe
        labels = df[self._dependent_variable]

        if labels.isna().sum().sum() != 0:
            raise Exception("You have some missing values in dependent variable column.")

        unique_labels = labels.unique()

        labels = np.array(labels)

        from numbers import Integral
        if isinstance(labels[0], (float, np.float32)) or len(unique_labels) > 20:
            return False

        if isinstance(int(labels[0]), (str, Integral)):
            return True

    def _is_categorical(self, labels):
        unique_labels = labels.unique()

        labels = np.array(labels)

        from numbers import Integral
        if isinstance(labels[0], (float, np.float32)) or len(unique_labels) > 20:
            return False

        if isinstance(int(labels[0]), (str, Integral)):
            return True

    @property
    def _databunch(self):
        if self._is_empty:
            return None

        return TabularDataObject._prepare_databunch(
            self._dataframe,
            self._field_mapping,
            self._procs,
            self._validation_indexes,
            self._bs
        )

    @property
    def _ml_data(self):
        if self._is_empty:
            return None, None, None, None

        if not HAS_NUMPY:
            raise Exception("This module requires numpy.")

        if not HAS_SK_LEARN:
            raise Exception("This module requires scikit-learn.")

        dataframe = self._dataframe

        labels = None

        if self._dependent_variable:
            labels = np.array(dataframe[self._dependent_variable])
            dataframe = dataframe.drop(self._dependent_variable, axis=1)

        if not self._procs:
            numerical_transformer = make_pipeline(
                SimpleImputer(strategy='median'),
                Normalizer())

            categorical_transformer = make_pipeline(
                SimpleImputer(strategy='constant')
            )

            _procs = make_column_transformer(
                (numerical_transformer, self._continuous_variables),
                (categorical_transformer, self._categorical_variables))
        else:
            _procs = self._procs

        self._encoder_mapping = None
        if self._categorical_variables:
            mapping = {}
            for variable in self._categorical_variables:
                labelEncoder = LabelEncoder()
                dataframe[variable] = np.array(labelEncoder.fit_transform(dataframe[variable]), dtype='int64')
                mapping[variable] = labelEncoder
            self._encoder_mapping = mapping

        processed_data = _procs.fit_transform(dataframe)

        training_data = processed_data.take(self._training_indexes, axis=0)
        training_labels = None
        if self._dependent_variable:
            training_labels = labels.take(self._training_indexes)

        validation_data = processed_data.take(self._validation_indexes, axis=0)
        validation_labels = None
        if self._dependent_variable:
            validation_labels = labels.take(self._validation_indexes)

        return training_data, training_labels, validation_data, validation_labels

    def _time_series_bunch(self, seq_len, normalize=True, bunch=True):
        if self._index_data is not None:
            bunched = []
            for i in range(len(self._index_data) - seq_len - 1):
                bunched.append(list(self._index_data[i:i + seq_len]))

            self._index_seq = np.array(bunched)

        if self._is_raster_only:
            return self._raster_timeseries_bunch(normalize, bunch)

        if len(list(self._dataframe.columns.values)) == 1:
            return self._univariate_bunch(seq_len, normalize, bunch)
        else:
            return self._multivariate_bunch(seq_len, normalize, bunch)

    def _raster_timeseries_bunch(self, normalize=True, bunched=True):
        kwargs_variables = {'num_workers': 0} if sys.platform == 'win32' else {}

        kwargs_variables['bs'] = self._bs

        if hasattr(arcgis, "env") and getattr(arcgis.env, "_processorType", "") == "CPU":
            kwargs_variables["device"] = torch.device('cpu')

        self._encoder_mapping = None
        mapping = {}
        df = self._dataframe.copy()  # .drop(self._dependent_variable, axis=1)

        for col in list(df.columns.values):
            if self._is_categorical(df[col]):
                labelEncoder = LabelEncoder()
                df[col] = np.array(labelEncoder.fit_transform(df[col]), dtype='int64')
                mapping[col] = labelEncoder

        self._encoder_mapping = mapping

        if normalize:
            if len(self._column_transforms_mapping) == 0:
                for col in list(df.columns):
                    self._column_transforms_mapping[col] = [MinMaxScaler()]
            else:
                for col in list(df.columns):
                    if len(self._column_transforms_mapping.get(col, [])) == 0:
                        self._column_transforms_mapping[col] = [DummyTransform()]

            processed_dataframe = df.copy()
            for col in list(df.columns):
                transformed_data = df[col]
                for transform in self._column_transforms_mapping.get(col, []):
                    transformed_data = transform.fit_transform(
                        np.array(transformed_data, dtype=df[col].dtype).reshape(-1, 1))
                    transformed_data = transformed_data.squeeze(1)
                processed_dataframe[col] = np.array(transformed_data, dtype=df[col].dtype)
        else:
            processed_dataframe = df.copy()

        big_bunch = []

        proc_df = processed_dataframe.copy()
        proc_df = proc_df.drop(self._dependent_variable, axis=1)

        for i in range(len(processed_dataframe)):
            big_bunch.append([proc_df.iloc[i].values])

        big_bunch = np.array(big_bunch)

        random.seed(self._seed)
        validation_indexes = random.sample(range(big_bunch.shape[0]),
                                           round(self._val_split_pct * big_bunch.shape[0]))
        self._validation_indexes_ts = validation_indexes

        self._training_indexes_ts = list(
            set([i for i in range(big_bunch.shape[0])]) - set(validation_indexes))

        X_train = big_bunch.take(self._training_indexes_ts, axis=0)
        X_valid = big_bunch.take(self._validation_indexes_ts, axis=0)

        y_train = np.array(processed_dataframe[self._dependent_variable].take(self._training_indexes_ts))
        y_valid = np.array(processed_dataframe[self._dependent_variable].take(self._validation_indexes_ts))

        if bunched is False:
            return X_train, X_valid, y_train, y_valid

        data = (ItemLists('.', TimeSeriesList(X_train), TimeSeriesList(X_valid))
                .label_from_lists(y_train, y_valid, label_cls=FloatList)
                .databunch(**kwargs_variables))

        return data

    def _multivariate_bunch(self, seq_len, normalize=True, bunched=True):
        kwargs_variables = {'num_workers': 0} if sys.platform == 'win32' else {}

        kwargs_variables['bs'] = self._bs

        if hasattr(arcgis, "env") and getattr(arcgis.env, "_processorType", "") == "CPU":
            kwargs_variables["device"] = torch.device('cpu')

        self._encoder_mapping = None
        mapping = {}
        df = self._dataframe.copy()#.drop(self._dependent_variable, axis=1)

        for col in list(df.columns.values):
            if self._is_categorical(df[col]):
                labelEncoder = LabelEncoder()
                df[col] = np.array(labelEncoder.fit_transform(df[col]), dtype='int64')
                mapping[col] = labelEncoder

        self._encoder_mapping = mapping

        if normalize:
            if len(self._column_transforms_mapping) == 0:
                for col in list(df.columns):
                    self._column_transforms_mapping[col] = [MinMaxScaler()]
            else:
                for col in list(df.columns):
                    if len(self._column_transforms_mapping.get(col, [])) == 0:
                        self._column_transforms_mapping[col] = [DummyTransform()]

            processed_dataframe = df.copy()
            for col in list(df.columns):
                transformed_data = df[col]
                for transform in self._column_transforms_mapping.get(col, []):
                    transformed_data = transform.fit_transform(np.array(transformed_data, dtype=df[col].dtype).reshape(-1, 1))
                    transformed_data = transformed_data.squeeze(1)
                processed_dataframe[col] = np.array(transformed_data, dtype=df[col].dtype)
        else:
            processed_dataframe = df.copy()

        big_bunch = []

        for i in range(len(processed_dataframe) - seq_len-1):
            bunch = []
            for col in list(processed_dataframe.columns.values):
                bunch.append(list(processed_dataframe[col][i:i + seq_len]))

            big_bunch.append(bunch)

        big_bunch = np.array(big_bunch)

        random.seed(self._seed)
        validation_indexes = random.sample(range(big_bunch.shape[0]),
                                           round(self._val_split_pct * big_bunch.shape[0]))
        self._validation_indexes_ts = validation_indexes

        self._training_indexes_ts = list(
            set([i for i in range(big_bunch.shape[0])]) - set(validation_indexes))

        X_train = big_bunch.take(self._training_indexes_ts, axis=0)
        X_valid = big_bunch.take(self._validation_indexes_ts, axis=0)

        y_train = np.array(processed_dataframe[self._dependent_variable].take(self._training_indexes_ts))
        y_valid = np.array(processed_dataframe[self._dependent_variable].take(self._validation_indexes_ts))

        if bunched is False:
            return X_train, X_valid, y_train, y_valid

        data = (ItemLists('.', TimeSeriesList(X_train), TimeSeriesList(X_valid))
                .label_from_lists(y_train, y_valid, label_cls=FloatList)
                .databunch(**kwargs_variables))

        return data

    def _univariate_bunch(self, seq_len, normalize=True, bunch=True):
        kwargs_variables = {'num_workers': 0} if sys.platform == 'win32' else {}

        kwargs_variables['bs'] = self._bs

        if hasattr(arcgis, "env") and getattr(arcgis.env, "_processorType", "") == "CPU":
            kwargs_variables["device"] = torch.device('cpu')

        df_columns = {}
        for i in range(seq_len):
            df_columns[f'att{i + 1}'] = []

        df_columns['target'] = []

        if self._is_classification:
            self._encoder_mapping = None
            mapping = {}
            labelEncoder = LabelEncoder()
            self._dataframe[self._dependent_variable] = np.array(labelEncoder.fit_transform(self._dataframe[self._dependent_variable]), dtype='int64')
            mapping[self._dependent_variable] = labelEncoder
            self._encoder_mapping = mapping

        if normalize:
            if not self._column_transforms_mapping.get(self._dependent_variable):
                self._column_transforms_mapping[self._dependent_variable] = [MinMaxScaler()]

            processed_dataframe = self._dataframe.copy()
            transformed_data = processed_dataframe[self._dependent_variable]
            for transform in self._column_transforms_mapping[self._dependent_variable]:
                transformed_data = transform.fit_transform(np.array(transformed_data, dtype=processed_dataframe[self._dependent_variable].dtype).reshape(-1, 1))
                transformed_data = transformed_data.squeeze(1)

            processed_dataframe[self._dependent_variable] = np.array(transformed_data, dtype=self._dataframe[self._dependent_variable].dtype)
        else:
            processed_dataframe = self._dataframe.copy()

        for i in range(len(processed_dataframe[self._dependent_variable]) - seq_len):
            for j in range(seq_len):
                if len(processed_dataframe[self._dependent_variable]) > i + seq_len - 1:
                    df_columns[f'att{j + 1}'].append(processed_dataframe[self._dependent_variable][i + j])
                else:
                    continue

            df_columns['target'].append(processed_dataframe[self._dependent_variable][i + seq_len])

        import pandas as pd
        df = pd.DataFrame(df_columns)

        columns = list(df.columns.values)
        columns.remove('target')

        random.seed(self._seed)
        validation_indexes = random.sample(range(len(df)),
                                           round(self._val_split_pct * len(df)))
        self._validation_indexes_ts = validation_indexes

        self._training_indexes_ts = list(
            set([i for i in range(len(df))]) - set(validation_indexes))

        y_train = np.array(df['target'].take(self._training_indexes_ts))
        y_valid = np.array(df['target'].take(self._validation_indexes_ts))

        X_train = To3dTensor(df.iloc[:, :-1].take(self._training_indexes_ts).values.astype(np.float32))
        X_valid = To3dTensor(df.iloc[:, :-1].take(self._validation_indexes_ts).values.astype(np.float32))

        if bunch is False:
            return X_train, X_valid, y_train, y_valid

        if self._is_classification:
            label_cls = CategoryList
        else:
            label_cls = FloatList

        data = (ItemLists('.', TimeSeriesList(X_train), TimeSeriesList(X_valid))
              .label_from_lists(y_train, y_valid, label_cls=label_cls)
              .databunch(**kwargs_variables)
              )

        return data

    def _process_data(self, dataframe):
        if not HAS_NUMPY:
            raise Exception("This module requires numpy.")

        if not HAS_SK_LEARN:
            raise Exception("This module requires scikit-learn.")

        if not self._procs:
            numerical_transformer = make_pipeline(
                SimpleImputer(strategy='median'),
                Normalizer())

            categorical_transformer = make_pipeline(
                SimpleImputer(strategy='constant')
            )

            _procs = make_column_transformer(
                (numerical_transformer, self._continuous_variables),
                (categorical_transformer, self._categorical_variables))
        else:
            _procs = self._procs

        if self._encoder_mapping:
            for variable, encoder in self._encoder_mapping.items():
                dataframe[variable] = np.array(encoder.fit_transform(dataframe[variable]), dtype='int64')

        processed_data = _procs.fit_transform(dataframe)

        return processed_data

    def show_batch(self, rows=5, graph=False, seq_len=None):
        """
        Shows a batch of dataframe prepared without applying transforms.
        In case of raster only, seq_len = number of rasters
        """

        if seq_len is not None or graph is True:
            self._show_graph(seq_len=seq_len, rows=rows)
            return

        if not rows or rows <= 0:
            rows = self._bs
        elif rows > len(self._training_indexes):
            rows = len(self._training_indexes)

        random_batch = random.sample(self._training_indexes, rows)
        return self._dataframe.loc[random_batch].sort_index()

    def _show_graph(self, seq_len=None, rows=5):
        """
        Shows a batch of prepared data in the form of graphs
        """

        if self._is_unsupervised:
            raise Exception("Show Graphs is used for Time Series Network")

        import matplotlib.pyplot as plt

        if seq_len is not None:
            X_train, X_valid, y_train, y_valid = self._time_series_bunch(seq_len, False, False)

            n_items = rows ** 2
            if n_items > len(X_train):
                n_items = len(X_train)

            sample = random.sample(range(len(X_train)), n_items)
            X_train_sample = np.array(X_train).take(sample, axis=0)
            y_train_sample = np.array(y_train).take(sample, axis=0)

            batched_index = []
            if self._index_data is not None:
                indexes = 0
                while indexes < len(self._index_data):
                    batched_index.append(self._index_data[indexes: indexes+seq_len])
                    indexes = indexes + seq_len
            else:
                j = 0
                while j < n_items:
                    batched_index.append([i for i in range(seq_len)])
                    j = j + 1

            rows = int(math.sqrt(n_items))

            fig, axs = plt.subplots(rows, rows, figsize=(10, 10))

            for i in range(rows):
                for j in range(rows):
                    for predictor in X_train_sample[i+j]:
                        axs[i, j].plot(batched_index[i+j], predictor)
                    axs[i, j].set_title(y_train_sample[i+j])
                    axs[i, j].tick_params(axis="x", labelrotation=60)

            plt.tight_layout()
            plt.show()
        else:
            # plotting the points
            # y = self._dataframe[self._dependent_variable]
            x_field = 'Time'
            if self._index_data is not None:
                x = self._index_data
                x_field = self._index_field
            else:
                x = [i for i in range(len(self._dataframe[self._dependent_variable]))]

            plt.figure(figsize=(25, 5))

            for col in list(self._dataframe.columns):
                plt.plot(x, self._dataframe[col], label=col)

            # naming the x axis
            plt.xlabel(x_field)
            # naming the y axis
            plt.ylabel(self._dependent_variable)
            plt.tick_params(axis="x", labelrotation=85)
            # giving a title to my graph
            plt.title('Data')
            plt.legend(loc='upper right')
            # function to show the plot
            plt.show()

    @staticmethod
    def _prepare_dataframe_from_features(
            input_features,
            dependent_variable,
            feature_variables=None,
            raster_variables=None,
            date_field=None,
            distance_feature_layers=None,
            index_field=None
    ):
        feature_variables = feature_variables if feature_variables else []
        raster_variables = raster_variables if raster_variables else []
        distance_feature_layers = distance_feature_layers if distance_feature_layers else []

        continuous_variables = []
        categorical_variables = []
        for field in feature_variables:
            if isinstance(field, tuple):
                if field[1]:
                    categorical_variables.append(field[0])
                else:
                    continuous_variables.append(field[0])
            else:
                continuous_variables.append(field)

        rasters = []
        for raster in raster_variables:
            if isinstance(raster, tuple):
                rasters.append(raster[0])
                if raster[1]:
                    band_count = raster[0].band_count
                    for index in range(band_count):
                        if index == 0:
                            categorical_variables.append(raster[0].name)
                        else:
                            categorical_variables.append(raster[0].name + f'_{index}')
                else:
                    band_count = raster[0].band_count
                    for index in range(band_count):
                        if index == 0:
                            continuous_variables.append(raster[0].name)
                        else:
                            continuous_variables.append(raster[0].name + f'_{index}')
            else:
                rasters.append(raster)
                band_count = raster.band_count
                for index in range(band_count):
                    if index == 0:
                        continuous_variables.append(raster.name)
                    else:
                        continuous_variables.append(raster.name + f'_{index}')

        dataframe, index_data = TabularDataObject._process_layer(
            input_features,
            date_field,
            distance_feature_layers,
            raster_variables,
            index_field
        )

        dataframe_columns = dataframe.columns
        if distance_feature_layers:
            count = 1
            while f'NEAR_DIST_{count}' in dataframe_columns:
                continuous_variables.append(f'NEAR_DIST_{count}')
                count = count + 1

        fields_to_keep = continuous_variables + categorical_variables
        if dependent_variable:
            fields_to_keep = fields_to_keep + [dependent_variable]

        for column in dataframe_columns:
            if column not in fields_to_keep:
                dataframe = dataframe.drop(column, axis=1)
            elif dependent_variable and column == dependent_variable:
                continue
            elif column in categorical_variables and dataframe[column].dtype == float:
                warnings.warn(f"Changing column {column} to continuous")
                categorical_variables.remove(column)
                continuous_variables.append(column)
            elif column in categorical_variables and dataframe[column].unique().shape[0] > 20:
                warnings.warn(f"Column {column} has more than 20 unique value. Sure this is categorical?")

        if date_field:
            date_fields = [
                ('Year', True), ('Month', True), ('Week', True),
                ('Day', True), ('Dayofweek', True), ('Dayofyear', False),
                ('Is_month_end', True), ('Is_month_start', True),
                ('Is_quarter_end', True), ('Is_quarter_start', True),
                ('Is_year_end', True), ('Is_year_start', True),
                ('Hour', True), ('Minute', True), ('Second', True), ('Elapsed', False)]

            for field in date_fields:
                if field[0] in dataframe_columns:
                    if field[1]:
                        categorical_variables.append(field[0])
                    else:
                        continuous_variables.append(field[1])

        return dataframe, {'dependent_variable': dependent_variable,
                           'categorical_variables': categorical_variables if categorical_variables else [],
                           'continuous_variables': continuous_variables if continuous_variables else [],
                           'index_data': index_data}

    @staticmethod
    def _process_layer(input_features, date_field, distance_layers, rasters, index_field):
        index_data = None
        if input_features is not None:
            if isinstance(input_features, FeatureLayer):
                input_layer = input_features
                sdf = input_features.query().sdf
            else:
                sdf = input_features.copy()
                input_layer = None
                try:
                    input_layer = sdf.spatial.to_feature_collection()
                except:
                    warnings.warn("Dataframe is not spatial, Rasters and distance layers will not work")

            if input_layer is not None and distance_layers:
                # Use proximity tool
                print("Calculating Distances.")
                count = 1
                for distance_layer in distance_layers:
                    output = arcgis.features.use_proximity.find_nearest(input_layer, distance_layer, max_count=1)
                    connecting_df = output['connecting_lines_layer'].query().sdf
                    near_dist = []

                    for i in range(len(connecting_df)):
                        near_dist.append(connecting_df.iloc[i]['Total_Miles'])

                    sdf[f'NEAR_DIST_{count}'] = near_dist
                    count = count + 1

            # Process Raster Data to get information.
            rasters_data = {}

            if input_layer is not None:
                original_points = []
                for i in range(len(sdf)):
                    original_points.append(sdf.iloc[i]["SHAPE"])

                input_layer_spatial_reference = sdf.spatial._sr
                for raster in rasters:
                    raster_type = 0

                    raster_calc = TabularDataObject._mean_of

                    if isinstance(raster, tuple):
                        if isinstance(raster[1], bool):
                            if raster[1] is True:
                                raster_type = 1
                                raster_calc = TabularDataObject._majority_of
                            if len(raster) > 2:
                                raster_calc = TabularDataObject._get_calc(raster_type, raster[2])
                        else:
                            raster_calc = TabularDataObject._get_calc(raster_type, raster[1])

                        raster = raster[0]

                    for i in range(raster.band_count):
                        if i == 0:
                            rasters_data[raster.name] = []
                        else:
                            rasters_data[raster.name + f'_{i}'] = []

                    shape_objects_transformed = arcgis.geometry.project(original_points, input_layer_spatial_reference,
                                                                        raster.extent['spatialReference'])
                    for shape in shape_objects_transformed:
                        shape['spatialReference'] = raster.extent['spatialReference']
                        if isinstance(shape, arcgis.geometry._types.Point):
                            raster_value = raster.read(origin_coordinate=(shape['x'], shape['y']), ncols=1, nrows=1)
                            value = raster_value[0][0]
                        elif isinstance(shape, arcgis.geometry._types.Polygon):
                            xmin, ymin, xmax, ymax = shape.extent
                            start_x, start_y = xmin + (raster.mean_cell_width / 2), ymin + (raster.mean_cell_height / 2)
                            values = []
                            while start_y < ymax:
                                while start_x < xmax:
                                    if shape.contains(arcgis.geometry._types.Point(
                                            {'x': start_x, 'y': start_y, 'sr': raster.extent['spatialReference']})):
                                        raster_read = raster.read(origin_coordinate=(start_x - raster.mean_cell_width, start_y),
                                                    ncols=1, nrows=1)[0][0]
                                        if len(values) == 0:
                                            for band in raster_read:
                                                values.append([band])
                                        else:
                                            index = 0
                                            for band_value in raster_read:
                                                values[index].append(band_value)
                                                index = index + 1

                                    start_x = start_x + raster.mean_cell_width
                                start_y = start_y + raster.mean_cell_height
                                start_x = xmin + (raster.mean_cell_width / 2)

                            if len(values) == 0:
                                raster_read = raster.read(origin_coordinate=(shape.true_centroid['x'] - raster.mean_cell_width, shape.true_centroid['y']), ncols=1,
                                                nrows=1)[0][0]
                                for band_value in raster_read:
                                    values.append([band_value])

                            value = raster_calc(values)
                        else:
                            raise Exception("Input features can be point or polygon only.")

                        for i in range(len(value)):
                            if i == 0:
                                rasters_data[raster.name].append(value[i])
                            else:
                                rasters_data[raster.name + f'_{i}'].append(value[i])

            # Append Raster data to sdf
            for key, value in rasters_data.items():
                sdf[key] = value
        else:
            try:
                import arcpy
            except:
                raise Exception("This function requires arcpy.")

            try:
                import pandas as pd
            except:
                raise Exception("This function requires pandas.")

            raster = rasters[0]
            if isinstance(raster, tuple):
                raster = raster[0]

            try:
                arcpy.env.outputCoordinateSystem = raster.extent['spatialReference']['wkt']
            except:
                arcpy.env.outputCoordinateSystem = raster.extent['spatialReference']['wkid']

            xmin = raster.extent['xmin']
            xmax = raster.extent['xmax']
            ymin = raster.extent['ymin']
            ymax = raster.extent['ymax']
            min_cell_size_x = raster.mean_cell_width
            min_cell_size_y = raster.mean_cell_height

            default_sr = raster.extent['spatialReference']

            for raster in rasters:
                if isinstance(raster, tuple):
                    raster = raster[0]

                point_upper = arcgis.geometry.Point(
                    {'x': raster.extent['xmin'], 'y': raster.extent['ymax'], 'sr': raster.extent['spatialReference']})
                point_lower = arcgis.geometry.Point(
                    {'x': raster.extent['xmax'], 'y': raster.extent['ymin'], 'sr': raster.extent['spatialReference']})
                cell_size = arcgis.geometry.Point(
                    {'x': raster.mean_cell_width, 'y': raster.mean_cell_height,
                     'sr': raster.extent['spatialReference']})

                points = arcgis.geometry.project([point_upper, point_lower, cell_size],
                                                 raster.extent['spatialReference'],
                                                 default_sr)
                point_upper = points[0]
                point_lower = points[1]
                cell_size = points[2]

                if xmin > point_upper.x:
                    xmin = point_upper.x
                if ymax < point_upper.y:
                    ymax = point_upper.y
                if xmax < point_lower.x:
                    xmax = point_lower.x
                if ymin > point_lower.y:
                    ymin = point_lower.y

                if min_cell_size_x > cell_size.x:
                    min_cell_size_x = cell_size.x

                if min_cell_size_y > cell_size.y:
                    min_cell_size_y = cell_size.y

            max_raster_columns = int(abs(math.ceil((xmax - xmin) / min_cell_size_x)))
            max_raster_rows = int(abs(math.ceil((ymax - ymin) / min_cell_size_y)))

            point_upper = arcgis.geometry.Point({'x': xmin, 'y': ymax, 'sr': default_sr})
            cell_size = arcgis.geometry.Point({'x': min_cell_size_x, 'y': min_cell_size_y, 'sr': default_sr})

            raster_data = {}
            for raster in rasters:
                if isinstance(raster, tuple):
                    raster = raster[0]
                field_name = raster.name
                point_upper_translated = \
                arcgis.geometry.project([point_upper], default_sr, raster.extent['spatialReference'])[0]
                cell_size_translated = \
                arcgis.geometry.project([cell_size], default_sr, raster.extent['spatialReference'])[0]
                raster_read = raster.read(
                    origin_coordinate=(point_upper_translated.x, point_upper_translated.y),
                    ncols=max_raster_columns, nrows=max_raster_rows,
                    cell_size=(cell_size_translated.x, cell_size_translated.y))

                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f'_{index}'
                            if not raster_data.get(key):
                                raster_data[key] = []

                            index = index + 1

                            raster_data[key].append(value)

            sdf = pd.DataFrame.from_dict(raster_data)

        if date_field:
            try:
                add_datepart(sdf, date_field)
            except:
                pass

        if index_field in list(sdf.columns.values):
            index_data = sdf[index_field].values

        return sdf, index_data

    @staticmethod
    def _prepare_databunch(
        dataframe,
        fields_mapping,
        procs=None,
        validation_indexes=[],
        batch_size=64
    ):

        if procs is None:
            procs = [Categorify, Normalize]
            fm = FillMissing(fields_mapping['categorical_variables'], fields_mapping['continuous_variables'])
            fm.add_col = False
            fm(dataframe)

        temp_file = tempfile.NamedTemporaryFile().name

        kwargs_variables = {'num_workers': 0} if sys.platform == 'win32' else {}

        kwargs_variables['cat_names'] = fields_mapping['categorical_variables']
        kwargs_variables['cont_names'] = fields_mapping['continuous_variables']
        kwargs_variables['bs'] = batch_size

        if hasattr(arcgis, "env") and getattr(arcgis.env, "_processorType", "") == "CPU":
            kwargs_variables["device"] = torch.device('cpu')

        data_bunch = TabularDataBunch.from_df(
            temp_file,
            dataframe,
            fields_mapping['dependent_variable'],
            procs=procs,
            valid_idx=validation_indexes,
            **kwargs_variables
        )

        return data_bunch

    @classmethod
    def _empty(cls, categorical_variables, continuous_variables, dependent_variable, encoder_mapping, procs=None):
        class_object = cls()
        class_object._dependent_variable = dependent_variable
        class_object._continuous_variables = continuous_variables
        class_object._categorical_variables = categorical_variables
        class_object._encoder_mapping = encoder_mapping
        class_object._is_empty = True
        class_object._procs = procs

        return class_object



import os
import json
import traceback
from pathlib import Path
from functools import partial
from ..models._arcgis_model import ArcGISModel, model_characteristics_folder
HAS_FASTAI = True

try:
    import torch
    import numpy as np
    import pandas as pd
    import torch.nn as nn
    from transformers import logging
    from fastai.train import to_fp16
    from fastprogress.fastprogress import progress_bar
    from fastai.basic_train import Learner, DatasetType
    from transformers import AutoConfig, AutoTokenizer, AutoModelForTokenClassification
    from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score, classification_report
    from .._utils.common import _get_emd_path
    from .._utils.text_transforms import get_results
    from .._utils.text_data import save_data_in_model_metrics_html, TextDataObject, copy_metrics
    from ._arcgis_transformer import ArcGISTransformer, ModelBackbone, infer_model_type, transformer_seq_length
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False
    transformer_seq_length = 512

    class ArcGISTransformer: pass


backbone_models_map = {
    'bert': ('bert-base-cased', 'bert-base-uncased', 'bert-large-cased', 'bert-large-uncased'),
    'albert': ('albert-base-v1', 'albert-base-v2', 'albert-large-v1', 'albert-large-v2',
               'albert-xlarge-v1', 'albert-xlarge-v2', 'albert-xxlarge-v1', 'albert-xxlarge-v2'),
    'roberta': ('roberta-base', 'roberta-large', 'distilroberta-base'),
    'distilbert': ('distilbert-base-cased', 'distilbert-base-uncased'),
    'xlnet': ('xlnet-base-cased', 'xlnet-large-cased'),
    'xlm': ('xlm-mlm-ende-1024', 'xlm-mlm-enfr-1024', 'xlm-mlm-xnli15-1024', 'xlm-mlm-en-2048'),
    'xlm-roberta': ('xlm-roberta-base', 'xlm-roberta-large'),
    'longformer': ('allenai/longformer-base-4096',),
    'electra': ('google/electra-base-discriminator', 'google/electra-base-generator'),
    'mobilebert': ('google/mobilebert-uncased',),
    'camembert': ('camembert-base',),
    'flaubert': ('flaubert/flaubert_small_cased', 'flaubert/flaubert_base_cased', 'flaubert/flaubert_base_uncased',
                 'flaubert/flaubert_large_cased')
}

transformer_architectures = ['BERT', 'RoBERTa', 'DistilBERT', 'ALBERT', 'CamemBERT', 'MobileBERT',
                             'XLNet', 'XLM', 'XLM-RoBERTa', 'FlauBERT', 'ELECTRA', 'Longformer']

backbone_models_reverse_map = {x:key for key, val in backbone_models_map.items() for x in val}


class TransformerNERLoss(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, output, labels, masks):
        # TransformerNER model itself outputs loss when forward method is called
        return output[0]


class TransformerForEntityRecognition(ArcGISTransformer):

    _supported_backbones = transformer_architectures

    def __init__(self, architecture, pretrained_model_name, config=None, pretrained_model_path=None,
                 seq_len=transformer_seq_length):
        super().__init__(architecture, pretrained_model_name, config, pretrained_model_path, task="ner")
        self._seq_len = seq_len
        self._max_seq_len = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % type(self).__name__

    def _process_config(self):
        """
        Process the model config to convert the class values to integers in the
        id2label attribute (which is a dictionary) of the model config
        """
        self._config.id2label = {int(x): y for x, y in self._config.id2label.items()}

    def _set_max_seq_length(self):
        """
        Set the max-seq-len parameter to avoid out of memory issues for transformers
        like Bart, Longformer, XLNet
        """
        self._max_seq_len = min(self._tokenizer.model_max_length, self._seq_len)

    @classmethod
    def _available_backbone_models(cls, architecture):
        """
        Provides a list of available models for a given transformer architecture

        =====================   =================================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------------
        architecture            Required string. The transformer architecture for
                                which we wish to get the available models
        ---------------------   -------------------------------------------------
        return: tuple containing available models for the given architecture
        """
        if architecture.lower() in backbone_models_map:
            return backbone_models_map[architecture.lower()]
        else:
            return f"Error, wrong architecture name - {architecture} supplied. " \
                   f"Please choose from {cls._supported_backbones}"

    def save(self, model_path):
        # Not required as this functionality will be implemented by the ArcGISModel child class
        raise NotImplementedError

    @classmethod
    def load_pretrained_model(cls, path):
        # Not required as this functionality will be implemented by the ArcGISModel child class
        raise NotImplementedError

    def forward(self, input_ids, attention_mask, token_type_ids, labels=None):
        if labels is not None:
            output = self._transformer(input_ids=input_ids, attention_mask=attention_mask,
                                      token_type_ids=token_type_ids, labels=labels)
        else:
            # Models like DistilBERT, RoBERTa, XLM-RoBERTa, Longformer, Bart, ELECTRA
            # and CamemBERT does not return token_type_ids in encode_plus method
            labels = token_type_ids
            output = self._transformer(input_ids=input_ids, attention_mask=attention_mask, labels=labels)
        return output

    def _load_transformer(self):
        """
        Method to load the transformer model. For this class we want to load the
        transformer model for performing Token Classification tasks
        """
        if self._pretrained_model_path and os.path.exists(self._pretrained_model_path):
            base_path, extension = os.path.splitext(self._pretrained_model_path)
            path = base_path + ".pth"
            if extension == ".emd" and os.path.exists(path):
                self._transformer = AutoModelForTokenClassification.from_pretrained(path, config=self._config)
            else:
                self._transformer = AutoModelForTokenClassification.from_pretrained(self._pretrained_model_path)
        else:
            self._transformer = AutoModelForTokenClassification.\
                from_pretrained(self._transformer_pretrained_model_name, config=self._config)

    @staticmethod
    def get_active_predictions_labels_per_batch(output, label, attention_mask):
        y_pred, y_true = [], []
        # import pdb; pdb.set_trace()
        predictions = output[1].argmax(2)
        active_tensors = attention_mask.view(-1) == 1

        predictions_flat = predictions.view(-1).tolist()
        labels_flat = label.view(-1).tolist()
        for index, item in enumerate(active_tensors.tolist()):
            if item:
                y_pred.append(predictions_flat[index])
                y_true.append(labels_flat[index])
        return y_pred, y_true

    @staticmethod
    def precision_score(output, label, attention_mask, average="micro"):
        y_pred, y_true = TransformerForEntityRecognition.\
            get_active_predictions_labels_per_batch(output, label, attention_mask)
        return torch.tensor(round(precision_score(y_true, y_pred, average=average, zero_division=1), 4))

    @staticmethod
    def recall_score(output, label, attention_mask, average="micro"):
        y_pred, y_true = TransformerForEntityRecognition. \
            get_active_predictions_labels_per_batch(output, label, attention_mask)
        return torch.tensor(round(recall_score(y_true, y_pred, average=average, zero_division=1), 4))

    @staticmethod
    def f1_score(output, label, attention_mask, average="micro"):
        y_pred, y_true = TransformerForEntityRecognition. \
            get_active_predictions_labels_per_batch(output, label, attention_mask)
        return torch.tensor(round(f1_score(y_true, y_pred, average=average, zero_division=1), 4))

    @staticmethod
    def get_active_predictions_labels(data_loader, learner, show_progress=True):
        predictions, labels = [], []
        dataset = list(iter(data_loader))
        if show_progress:
            for i in progress_bar(range(len(data_loader))):
                batch = dataset[i]
                output = learner.model.forward(*batch[0])
                batch_preds, batch_labels = TransformerForEntityRecognition.\
                    get_active_predictions_labels_per_batch(output, *batch[1])
                predictions.extend(batch_preds)
                labels.extend(batch_labels)
        else:
            results = [TransformerForEntityRecognition.get_active_predictions_labels_per_batch
                       (learner.model.forward(*batch[0]), *batch[1]) for batch in dataset]

            for item in results:
                predictions.extend(item[0])
                labels.extend(item[1])

        return predictions, labels

    def generate_inference(self, text_list, device):
        tokenizer = self._tokenizer
        max_seq_length = self._max_seq_len
        model_type = self._transformer_architecture
        token_list = [x.split(" ") for x in text_list]
        if model_type in ["roberta", "bart", "longformer"]:
            token_list = [[f" {y}" for y in x] for x in token_list]

        encodings = tokenizer.batch_encode_plus(token_list, max_length=max_seq_length, padding=True, truncation=True,
                                                is_split_into_words=True, return_tensors="pt").to(device)

        input_ids, attention_mask = encodings.get("input_ids"), encodings.get("attention_mask")
        token_type_ids = encodings.get("token_type_ids")
        # Models like DistilBERT, RoBERTa, XLM-RoBERTa, Longformer, Bart, ELECTRA
        # and CamemBERT does not return token_type_ids in encode_plus method
        if token_type_ids is not None:
            output = self._transformer(input_ids=input_ids, attention_mask=attention_mask,
                                       token_type_ids=token_type_ids)
        else:
            output = self._transformer(input_ids=input_ids, attention_mask=attention_mask)

        batch_tokens = input_ids.tolist()
        batch_labels = output[0].argmax(2).tolist()

        return batch_tokens, batch_labels


class _TransformerEntityRecognizer(ArcGISModel):

    supported_backbones = transformer_architectures

    def __init__(self, data, backbone="bert-base-cased", **kwargs):
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error
            _raise_fastai_import_error(import_exception=import_exception)

        self.logger = logging.get_logger()
        if kwargs.get('verbose', None):
            self.logger.setLevel(kwargs.get('verbose').upper())
        else:
            self.logger.setLevel(logging.ERROR)

        model_backbone = ModelBackbone(backbone)
        super().__init__(data, model_backbone)
        self.stats = "macro"
        self._is_empty = True
        self.train_ds, self.val_ds = None, None
        self._address_tag = self._data._address_tag
        self._mixed_precision = kwargs.get('mixed_precision', False)
        self._seq_len = kwargs.get('seq_len', transformer_seq_length)
        self._create_text_learner_object(
            data, backbone, kwargs.get('pretrained_path', None), mixed_precision=self._mixed_precision, seq_len=self._seq_len)

        self.learn.model = self.learn.model.to(self._device)
        layer_groups = self.learn.model.get_layer_groups()
        self.learn.split(layer_groups)
        # Freeze the model by default
        self._freeze()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % type(self).__name__.replace('_', '')

    @classmethod
    def available_backbone_models(cls, architecture):
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error
            _raise_fastai_import_error(import_exception=import_exception)
        return TransformerForEntityRecognition._available_backbone_models(architecture)

    def _create_text_learner_object(self, data, backbone, pretrained_path=None, mixed_precision=False,
                                    seq_len=transformer_seq_length):

        model_type = infer_model_type(backbone, transformer_architectures)
        self.logger.info(f"Inferred Backbone: {model_type}")
        pretrained_model_name = backbone

        transformer_tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)
        if data._is_empty or data._backbone != backbone:
            self.logger.info('Creating DataBunch')
            data._prepare_databunch(tokenizer=transformer_tokenizer, model_type=model_type,
                                    seq_len=seq_len, backbone=backbone, logger=self.logger)

        databunch = data.get_databunch()

        config = AutoConfig.from_pretrained(pretrained_model_name)
        config.label2id = data._label2id
        config.id2label = {y: x for x, y in config.label2id.items()}

        if pretrained_path is not None: pretrained_path = str(_get_emd_path(pretrained_path))

        model = TransformerForEntityRecognition(
            architecture=model_type,
            pretrained_model_name=pretrained_model_name,
            config=config,
            pretrained_model_path=pretrained_path,
            seq_len=seq_len
        )

        model.init_model()

        loss_func = TransformerNERLoss()

        f1 = partial(TransformerForEntityRecognition.f1_score, average=self.stats)
        recall = partial(TransformerForEntityRecognition.recall_score, average=self.stats)
        precision = partial(TransformerForEntityRecognition.precision_score, average=self.stats)
        metrics = [precision, recall, f1]

        self.learn = Learner(databunch, model, loss_func=loss_func, metrics=metrics)

        if pretrained_path is not None:
            self.load(pretrained_path)

        if mixed_precision:
            if model_type in ["xlnet", "mobilebert"]:
                error_message = (f"Mixed precision training is not supported for transformer model - {model_type.upper()}."
                                 "\nKindly turn off the `mixed_precision` flag to use this model in its default mode,"
                                 f" or choose a different transformer architectures from - {transformer_architectures}")
                raise Exception(error_message)
            self.logger.info("Converting model to 16 Bit Floating Point precision")
            self.learn = to_fp16(self.learn)

        if databunch.is_empty:
            self.train_ds = None
            self.val_ds = None
            self._is_empty = True
        else:
            self.train_ds = databunch.train_ds
            self.valid_ds = databunch.valid_ds
            self._is_empty = False

        self.entities = data._unique_tags
        self._address_tag = data._address_tag

    def _freeze(self):
        """
        Freeze up to last layer group to train only the last layer group of the model.
        """
        self.learn.freeze()

    def _save_df_to_html(self, path):
        if getattr(self._data, '_is_empty', False):
            copy_metrics(self._data.emd_path, path, model_characteristics_folder)
            return

        metrics_per_label = self.metrics_per_label(show_progress=False)
        metrics_per_label_str = metrics_per_label.to_html().replace(">\n", ">")

        dataframe = self.show_results()
        show_result_df_str = dataframe.to_html(index=False, justify='left').replace(">\n", ">")

        show_result_msg = "<p><b>Sample Results</b></p>"
        metrics_per_label_msg = "<p><b>Metrics per label</b></p>"

        text = f"\n\n{metrics_per_label_msg}\n\n{metrics_per_label_str}\n\n{show_result_msg}\n\n{show_result_df_str}"

        save_data_in_model_metrics_html(text, path, model_characteristics_folder)

    def save(self, name_or_path, framework='PyTorch', publish=False, gis=None, compute_metrics=True,
             save_optimizer=False, **kwargs):
        from ..models._arcgis_model import _create_zip
        zip_files = kwargs.pop('zip_files', True)
        overwrite = kwargs.pop('overwrite', False)
        path = super().save(name_or_path, framework, publish=False, gis=None, compute_metrics=compute_metrics,
                            save_optimizer=save_optimizer, zip_files=False, **kwargs)

        self._save_df_to_html(path)

        if zip_files:
            _create_zip(path.name, str(path))

        if publish:
            self._publish_dlpk((path/path.stem).with_suffix('.dlpk'), gis=gis, overwrite=overwrite)

        return Path(path)

    @property
    def _model_metrics(self):
        from IPython.utils import io
        with io.capture_output() as captured:
            metrics = self._calculate_model_metrics()
            per_class_metric_df = self.metrics_per_label()
            metrics["metrics_per_label"] = per_class_metric_df.transpose().to_dict()
        return {"Metrics": json.dumps(metrics)}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Architecture"]= self.learn.model._transformer_architecture
        _emd_template["PretrainedModel"]= self.learn.model._transformer_pretrained_model_name
        _emd_template["ModelType"] = "Transformer"
        _emd_template["MixedPrecisionTraining"] = self._mixed_precision
        _emd_template["AddressTag"] = self._address_tag
        _emd_template["Labels"] = list(self.learn.model._config.label2id.keys())
        _emd_template["Label2Id"] = self.learn.model._config.label2id
        _emd_template["SequenceLength"] = self._seq_len
        return _emd_template

    def load(self, name_or_path):
        """
        Loads a saved EntityRecognizer model from disk.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        if '\\' in str(name_or_path) or '/' in str(name_or_path):
            name_or_path = str(_get_emd_path(name_or_path))
        else:
            name_or_path = Path('models') / name_or_path
            name_or_path = str(_get_emd_path(name_or_path))
        return super().load(name_or_path)

    @classmethod
    def from_model(cls, emd_path, data=None):
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        mixed_precision = emd["MixedPrecisionTraining"]
        entities = set(emd["Labels"])
        label2id = emd["Label2Id"]
        address_tag = emd["AddressTag"]
        pretrained_model = emd["PretrainedModel"]
        seq_len = emd.get("SequenceLength", transformer_seq_length)
        data_is_none = False
        if data is None:
            data_is_none = True
            data = TextDataObject(task="ner")
            data.create_empty_object_for_ner(entities, address_tag, label2id)
            data.emd, data.emd_path = emd, emd_path.parent

        data._is_empty = True
        data._label2id = label2id
        cls_object = cls(data, pretrained_model, pretrained_path=str(emd_path), mixed_precision=mixed_precision, seq_len=seq_len)
        if data_is_none: cls_object._data._is_empty = True
        return cls_object

    def extract_entities(self, text_list, batch_size=4, drop=True, debug=False):
        results, columns, file_names = [], [], []
        if isinstance(text_list,  (str, bytes)):
            path = text_list
            text_list, skipped_docs = [], []
            item_names = os.listdir(path)
            for item_name in item_names:
                try:
                    with open(f'{path}/{item_name}', 'r', encoding='utf-16', errors='ignore') as f:
                        text_list.append(f.read())
                    file_names.append(item_name)
                except:
                    try:
                        with open(f'{path}/{item_name}', 'r', encoding='utf-8', errors='ignore') as f:
                            text_list.append(f.read())
                        file_names.append(item_name)
                    except Exception as e:
                        self.logger.exception(e)
                        skipped_docs.append(item_name)
            if len(skipped_docs):
                print('Unable to read the following documents ', ', '.join(skipped_docs))

        tokenizer, id2label = self.learn.model._tokenizer, self.learn.model._config.id2label
        model_type = self.learn.model._transformer_architecture
        self.logger.info(f"Generating Inference using - {model_type} transformer model.")
        for i in progress_bar(range(0, len(text_list), batch_size)):
            tokens, labels = self.learn.model.generate_inference(text_list[i: i + batch_size], self._device)
            if debug:
                batch_results = get_results(tokens, labels, tokenizer, id2label, model_type, len(tokens))
                results.extend(batch_results)
            else:
                batch_results, columns = self._process_results(tokens, labels, return_dataframe=False,
                            drop=drop, start_index=i, file_names=file_names[i: i + batch_size])
                results.extend(batch_results)

        if debug: return results
        dataframe = pd.DataFrame(results, columns=columns)
        dataframe.fillna("", inplace=True)
        return dataframe

    def _process_results(self, tokens, predictions, return_dataframe=True, drop=False, start_index=0, file_names=[]):
        data_list = []
        tokenizer = self.learn.model._tokenizer
        id2label = self.learn.model._config.id2label
        model_type = self.learn.model._transformer_architecture
        results = get_results(tokens, predictions, tokenizer, id2label, model_type, num_items=len(tokens))

        columns = self._data._unique_tags
        columns.discard("O")
        address_tag, text_tag = self._address_tag, "Text"
        has_address = True if address_tag in self.entities else False

        if has_address:
            cols = [x for x in columns if x not in [text_tag, address_tag]]
        else:
            cols = [x for x in columns if x not in [text_tag]]

        for index, row in enumerate(results):
            text = row[text_tag]
            if has_address:
                values = [", ".join(row.get(column, "")) for column in cols]
                address_list = row.get(address_tag)
                # Don't append the row if drop flag is True and address_list is None
                if drop is True and address_list is None:
                    continue
                if address_list is None: address_list = [""]

                file_name_column = file_names[index] if len(file_names) else f"Example_{index + start_index}"
                for address in address_list:
                    data_list.append([text, file_name_column, address, *values])
            else:
                file_name_column = file_names[index] if len(file_names) else f"Example_{index}"
                values = [", ".join(row.get(column, "")) for column in cols]
                data_list.append([text, file_name_column, *values])

        # data_list = data_list[:self._data._bs]

        df_columns = [text_tag, "Filename", address_tag, *cols] if has_address else [text_tag, "Filename", *cols]
        if return_dataframe:
            dataframe = pd.DataFrame(data_list, columns=df_columns)
            dataframe.fillna("", inplace=True)
            return dataframe
        else:
            return data_list, df_columns

    def show_results(self, ds_type='Valid'):
        """
        Runs entity extraction on a random batch from the mentioned ds_type.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        ds_type                 Optional string, defaults to valid.
        =====================   ===========================================

        :returns: Pandas DataFrame
        """
        self._check_requisites()
        databunch = self._data.get_databunch()
        x, y = databunch.one_batch(ds_type=ds_type, detach=False)
        output = self.learn.model.forward(*x)
        predictions = output[1].argmax(2).tolist()
        tokens = x[0].tolist()
        df = self._process_results(tokens, predictions)
        return df

    def _calculate_model_metrics(self, metric_type="all"):
        metrics = {}
        databunch = self._data.get_databunch()
        predictions, labels = TransformerForEntityRecognition. \
            get_active_predictions_labels(databunch.valid_dl, self.learn)

        if metric_type in ["precision_score", "all"]:
            metrics["precision_score"] = round(
                precision_score(labels, predictions, average=self.stats, zero_division=1), 2)

        if metric_type in ["recall_score", "all"]:
            metrics["recall_score"] = round(recall_score(labels, predictions, average=self.stats, zero_division=1), 2)

        if metric_type in ["f1_score", "all"]:
            metrics["f1_score"] = round(f1_score(labels, predictions, average=self.stats, zero_division=1), 2)

        if metric_type in ["accuracy_score", "all"]:
            metrics["accuracy"] = round(accuracy_score(labels, predictions), 2)

        return metrics

    def _get_metric(self, metric_type):
        try:
            self._check_requisites()
        except Exception as e:
            metrics = self._data.emd.get('Metrics')
            if metrics: return json.loads(metrics).get(metric_type)
            else: self.logger.error("Metric not found in the loaded model")
        else:
            if hasattr(self.learn, 'recorder'):
                metrics_names = self.learn.recorder.metrics_names
                metrics_values = self.learn.recorder.metrics
                if metric_type in self.learn.recorder.metrics_names:
                    index = metrics_names.index(metric_type)
                    return round(metrics_values[-1][index].item(), 2)
                else:
                    metrics = self._calculate_model_metrics(metric_type)
                    return metrics[metric_type]
            else:
                metrics = self._calculate_model_metrics(metric_type)
                return metrics[metric_type]

    def precision_score(self):
        return self._get_metric("precision_score")

    def recall_score(self):
        return self._get_metric("recall_score")

    def f1_score(self):
        return self._get_metric("f1_score")

    def metrics_per_label(self, show_progress=True):
        try:
            self._check_requisites()
        except Exception as e:
            metrics = self._data.emd.get('Metrics')
            if metrics:
                per_label_metrics = json.loads(metrics).get("metrics_per_label", {})
                return self._create_dataframe_from_dict(per_label_metrics)
            else: self.logger.error("Metric not found in the loaded model")
        else:
            databunch = self._data.get_databunch()
            predictions, labels = TransformerForEntityRecognition.\
                get_active_predictions_labels(databunch.valid_dl, self.learn, show_progress=show_progress)
            target_names = databunch.train_ds.label2id
            output = classification_report(labels, predictions, target_names=target_names, zero_division=0, output_dict=True)
            return self._create_dataframe_from_dict(output)

    @staticmethod
    def _create_dataframe_from_dict(out_dict):
        out_dict.pop("accuracy", None)
        out_dict.pop("macro avg", None)
        out_dict.pop("weighted avg", None)
        df = pd.DataFrame(out_dict)
        df.drop("support", inplace=True, errors="ignore")
        dataframe = df.T.round(2)
        column_mappings = {'precision': 'Precision_score', 'recall': 'Recall_score', 'f1-score': 'F1_score'}
        dataframe.rename(columns=column_mappings, inplace=True)
        return dataframe

    def plot_losses(self, show=True):
        """
        Plot training and validation losses.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        show                    Optional bool. Defaults to True
                                If set to False, figure will not be plotted
                                but will be returned, when set to True function
                                will plot the figure and return nothing.
        =====================   ===========================================

        :returns: matplotlib.figure.Figure
        """
        self._check_requisites()
        import matplotlib.pyplot as plt
        if not hasattr(self.learn, 'recorder'):  # return none if the recorder is empty
            self.logger.error("Model needs to be trained first. Please call `model.fit()` to train the model."
                         " Then call this method to plot/return the loss curve.")
            return
        return_fig = not show
        fig = self.learn.recorder.plot_losses(return_fig=return_fig)
        if return_fig:
            plt.close()
            return fig
